/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc> 

    This file is part of opkg2json.

    dsplwriter is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    opkg2json is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with opkg2json. If not, see <http://www.gnu.org/licenses/>.

*/

#include "autoconfig.h"

// workaround for broken opkg headers
#define OPKG2JSON_VERSION VERSION
#undef VERSION
#undef PACKAGE_VERSION
#undef PACKAGE_TARNAME
#undef PACKAGE
#undef PACKAGE_BUGREPORT
#undef PACKAGE_STRING
#undef PACKAGE_NAME

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <opkg.h>
#include <pkg_parse.h>
#include <json.h>

#if defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)
    #include <getopt.h>
    #define OPTSTR "liutf:"
#else
    #error Need getopt_long!
#endif

#define ACTION_TYPE_LIST            1
#define ACTION_TYPE_LIST_INSTALLED  2
#define ACTION_TYPE_LIST_UPGRADABLE 3
#define ACTION_TYPE_UPGRADE_SIZE    4

typedef struct
{
    int action;
    json_object *json_array;
    int size;
    int list_tags;
    char *filter;
} user_data_t;

static void print_copyright()
{
    printf("\nopkg2json version %s\n", OPKG2JSON_VERSION);
    printf("Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland\n\n");
}

static void msg(int level, const char *fmt, va_list ap)
{
    vfprintf(stderr, fmt, ap);
}

static void package_to_json(pkg_t *pkg, void *user_data) 
{
    json_object *json_pkg;
    json_object *json_pkg_array;

    int free_version = 1;

    if ((pkg == NULL) || (pkg->name == NULL))
    {
        fprintf(stderr, "libopkg returned invalid package data\n");
        return;
    }

    if (strlen(pkg->name) >= 4)
    {
        if (strstr(pkg->name + (strlen(pkg->name) - 4), "-dbg") ||
            strstr(pkg->name + (strlen(pkg->name) - 4), "-dev"))
        {
            return;
        }
    }

    if (user_data == NULL)
    {
        fprintf(stderr, "libopkg returned invalid user data\n");
        return;
    }

    if (((user_data_t *)user_data)->filter != NULL)
    {
        size_t l = strlen(((user_data_t *)user_data)->filter);
        if (strncmp(pkg->name, ((user_data_t *)user_data)->filter, l) != 0)
        {
            return;
        }
    }

    int action = ((user_data_t *)user_data)->action;
    if ((action == ACTION_TYPE_LIST_INSTALLED) && 
        (pkg->state_status != SS_INSTALLED))
    {
        return;
    } 
    else if ((action == ACTION_TYPE_LIST) && 
             (pkg->state_status == SS_INSTALLED))
    {
        return;
    }

    if ((((user_data_t *)user_data)->action == ACTION_TYPE_LIST_UPGRADABLE) ||
        (((user_data_t *)user_data)->action == ACTION_TYPE_UPGRADE_SIZE))
    {
        ((user_data_t *)user_data)->size += pkg->size;
    }

    if (((user_data_t *)user_data)->action == ACTION_TYPE_UPGRADE_SIZE)
    {
        return;
    }

    char *v = pkg_version_str_alloc(pkg);
    if (v == NULL)
    {
        if (pkg->version != NULL)
            v = pkg->version;
        else 
        {
            v = "";
            free_version = 0;
        }
    }

    json_pkg_array = ((user_data_t *)user_data)->json_array;

    json_pkg = json_object_new_object();
    if (pkg->name)
    {
        json_object_object_add(json_pkg, "name", 
                               json_object_new_string(pkg->name));
    }
    else
    {
        json_object_object_add(json_pkg, "name", json_object_new_string("")); 
    }

    json_object_object_add(json_pkg, "version", 
                                json_object_new_string(v));

    if (pkg->maintainer)
    {
        json_object_object_add(json_pkg, "maintainer", 
                                json_object_new_string(pkg->maintainer));
    }
    else
    {
        json_object_object_add(json_pkg, "maintainer",
                                json_object_new_string(""));
    }

    if (pkg->description)
    {
        json_object_object_add(json_pkg, "description", 
                               json_object_new_string(pkg->description));
    }
    else
    {
        json_object_object_add(json_pkg, "description", 
                                json_object_new_string(""));
    }

    json_object_object_add(json_pkg, "state",
                           json_object_new_int(pkg->state_status));

    // when the page is not set the value will be "unknown", otherwise it
    // should be a valid url starting with http:// but we need to be fast
    // so won't check that
    if ((pkg->homepage != NULL) && (pkg->homepage[0] == 'h'))
    {
        json_object_object_add(json_pkg, "homepage",
                               json_object_new_string(pkg->homepage));
    }
    else
    {
        json_object_object_add(json_pkg, "homepage", 
                               json_object_new_string(""));
    }

    if (((user_data_t *)user_data)->list_tags == 1)
    {
        if (pkg->tags != NULL)
        {
            json_object_object_add(json_pkg, "tags",
                                   json_object_new_string(pkg->tags));
        }
        else
        {
            json_object_object_add(json_pkg, "tags",
                                   json_object_new_string(""));
        }
    }

    json_object_array_add(json_pkg_array, json_pkg);

    if (v && free_version) 
    {
        free(v);
    }
}


static void print_usage(const char *myname)
{
    print_copyright();

    fprintf(stderr, "\n%s options\n\
\n\
Supported options:\n\
    --list or -l                list packges available in feed\n\
    --list-installed or -i      list installed packages\n\
    --list-upgradable or -u     list upgradable packages\n\
    --tags or -t                add tags field to the listing\n\
    --upgrade-size or -s        show size of pending upgrade\n\
    --filter or -f              list only packages starting with \"string\"\n\
    --version                   print version information and exit\n\
    --help                      this help message\n\
\n", myname);
}

int main(int argc, char **argv)
{
    json_object *json_result;
    user_data_t user_data;

    int list_installed = 0;
    int list = 0;
    int list_upgradable = 0;
    int list_tags = 0;
    int upgrade_size = 0;

    int opt_index = 0;
    int o;
    user_data.filter = NULL;

    static struct option long_options[] = {
        {"list-installed",  0, 0, 'i'},
        {"list-upgradable", 0, 0, 'u'},
        {"tags",            0, 0, 't'},
        {"upgrade-size",    0, 0, 's'},
        {"list",            0, 0, 'l'},
        {"filter",          1, 0, 'f'}, 
        {"version",         0, 0, 'v'},
        {"help",            0, 0, 'h'},
        {0, 0, 0, 0}
    };

    while (1)
    {
        o = getopt_long(argc, argv, OPTSTR, long_options, &opt_index);
        if (o == -1) break;

        switch (o)
        {
            case 'l':
                list = 1;
                break;
            case 'i':
                list_installed = 1;
                break;
            case 'u':
                list_upgradable = 1;
                break;
            case 's':
                upgrade_size = 1;
                break;
            case 't':
                list_tags = 1;
                break;
            case 'f':
                user_data.filter = optarg;
                break;
            case 'v':
                print_copyright();
                exit(EXIT_SUCCESS);
                break;
            case '?':
            case 'h':
            default:
                print_usage(argv[0]);
                exit(EXIT_FAILURE);
        }
    }

    if (list && list_installed && list_upgradable && upgrade_size)
    {
        print_usage(argv[0]);
        exit(EXIT_FAILURE);
    }

    json_result = json_object_new_object();
    if (json_result == NULL)
    {
        fprintf(stderr, "Failed to initialize json object!\n");
        fprintf(stdout, "{ \"success\": false, \"code\": -1, \"message\": "
                        "\"Failed to initialize json object!\" }");
        exit(EXIT_FAILURE);
    }

    if (user_data.action != ACTION_TYPE_UPGRADE_SIZE)
    {
        user_data.json_array = json_object_new_array();
        if (user_data.json_array == NULL)
        {
            fprintf(stderr, "Failed to initialize json array object!\n");
            fprintf(stdout, "{ \"success\": false, \"code\": -1, \"message\": "
                            "\"Failed to initialize json array object!\" }");
            json_object_put(json_result);
            exit(EXIT_FAILURE);
        }
    }

    conf->pfm = PFM_SOURCE;
    conf->verbosity = NOTICE;
    conf->opkg_vmessage = msg;

    if (opkg_new() != 0)
    {
        fprintf(stderr, "Failed to initialize opkg!\n");
        fprintf(stdout, "{ \"success\": false, \"code\": -1, \"message\": "
                        "\"Failed to initialize opkg!\" }");
        json_object_put(json_result);
        json_object_put(user_data.json_array);
        exit(EXIT_FAILURE);
    }

    user_data.size = 0;
    user_data.list_tags = list_tags;

    if (list)
    {
        user_data.action = ACTION_TYPE_LIST;
        opkg_list_packages(package_to_json, &user_data);
    }
    else if (list_installed) 
    {
        user_data.action = ACTION_TYPE_LIST_INSTALLED;
        opkg_list_packages(package_to_json, &user_data);
    }
    else if (list_upgradable) 
    {
        user_data.action = ACTION_TYPE_LIST_UPGRADABLE;
        opkg_list_upgradable_packages(package_to_json, 1, &user_data);
    }
    else if (upgrade_size)
    {
        user_data.action = ACTION_TYPE_UPGRADE_SIZE;
        opkg_list_upgradable_packages(package_to_json, 1, &user_data);
    }
    else
    {
        print_usage(argv[0]);
        exit(EXIT_FAILURE);
    }

    json_object_object_add(json_result, "success", json_object_new_boolean(1));
    json_object_object_add(json_result, "code", json_object_new_int(0));
    json_object_object_add(json_result, "message",json_object_new_string("OK"));
    if (user_data.action != ACTION_TYPE_UPGRADE_SIZE)
    {
        json_object_object_add(json_result, "data", user_data.json_array);
    }

    if ((user_data.action == ACTION_TYPE_LIST_UPGRADABLE) ||
        (user_data.action == ACTION_TYPE_UPGRADE_SIZE))
    {
        json_object_object_add(json_result, "size",
                               json_object_new_int(user_data.size));
    }

    fprintf(stdout, json_object_to_json_string(json_result));
    
    json_object_put(json_result);

    opkg_free();

    return EXIT_SUCCESS;
}
